/* Copyright 2007-2013, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */
/* Modules:
 * Syscon -- DIO, LED, etc
 * CAN
 * XUARTs
 * ADC core
 * PC/104
 * SD card
 */
module ts7250_v2_top(

  // SMC bus
  inout [15:0] smc_d_pad,
  input smc_clk_pad,
  inout smc_rdy_pad,
  input smc_cs1n_pad,
  input smc_llan_pad,
  input smc_luan_pad,
  input smc_be1n_pad,
  input smc_be2n_pad,
  input smc_wen_pad,
  input smc_rdn_pad,
  output smc_irq_pad,

  // Serial ports and CAN
  output db9_txd_pad,
  output db9_rts_pad,
  output db9_dtr_pad,
  input db9_rxd_pad,
  input db9_cts_pad,
  input db9_dsr_pad,
  input db9_dcd_pad,
  output com2_txd_pad,
  output com2_rts_pad,
  input com2_rxd_pad,
  input com2_cts_pad,
  output com3_txd_pad,
  output com3_rts_pad,
  input com3_rxd_pad,
  input com3_cts_pad,
  output txd_485_pad,
  output tx_en_485_pad,
  input rxd_485_pad,
  input rxd_422_pad,
  output txd_can_pad,
  input rxd_can_pad,
  input debug_txd_pad,
  output debug_rxd_pad,

  // PC/104
  inout [15:0] isa_dat_pad,
  inout [19:0] isa_add_pad,
  inout isa_iorn_pad,
  inout isa_iown_pad,
  inout isa_memwn_pad,
  inout isa_memrn_pad,
  inout isa_reset_pad,
  inout isa_14mhz_pad,
  inout isa_endx_pad,
  inout isa_aen_pad,
  inout isa_refresh_pad,
  inout isa_iochk_pad,
  inout isa_iordy_pad,
  inout isa_drq3_pad,
  inout isa_drq2_pad,
  inout isa_dack3_pad,
  inout isa_irq5_pad,
  inout isa_irq6_pad,
  inout isa_irq7_pad,
  output isa_data_enn_pad,
  output isa_dir_pad,
  inout isa_irq9_pad,
  inout isa_dack1_pad,
  inout isa_drq1_pad,
  inout isa_irq3_pad,
  inout isa_bclk_pad,
  inout isa_dack2_pad,
  inout isa_tc_pad,
  inout isa_bale_pad,
  inout [15:0] isa_cn_d_pad,

  // Control lines
  input i2c_clk_pad,
  input i2c_dat_pad,
  input fpga_25mhz_pad,
  input jp1_pad,
  input jp2n_pad,
  input boot_flag,
  output m0_12mhz_clk_pad,
  output m0_rxd_pad,
  input m0_txd_pad,
  output en_usb_5v_pad,
  input cpu_resetn_pad, // now an FPGA reset input?
  output reboot_pad,
  inout red_ledn_pad,
  inout green_ledn_pad,
  inout eth_right_ledn_pad,

  // SD cards
  inout sd_d0_pad,
  inout sd_d1_pad,
  inout sd_d2_pad,
  inout sd_d3_pad,
  inout sd_cmd_pad,
  output sd_clk_mic_pad,
  output sd_clk_fs_pad,
  output sd_powern_pad,
  output fault_sd_ledn_pad,
  output sd_act_ledn_pad,
  input sd_wp_pad,

  // SPI
  output spi_mosi_pad,
  input spi_miso_pad,
  output spi_clk_pad,
  output spi_csn_pad,

  // MCP3425 ADC
  inout adc_dat_pad,
  inout adc_clk_pad,
  output [2:0] adc_a_pad,

  // eMMC flash
  output fault_emmc_ledn_pad,
  inout [7:0] emmc_d_pad,
  output emmc_clk_pad,
  inout emmc_cmd_pad,
  output emmc_resetn_pad,

  // DIO and LCD headers
  inout [8:0] dio_pad,
  output en_dio_fet_pad,
  inout lcd_en_pad,
  inout lcd_wrn_pad,
  inout lcd_rs_pad,
  inout lcd_bias_pad,
  inout [7:0] lcd_pad,
  inout m0_spare_pad,

  // I like to assign all pads so Lattice tools can't move them around
  output [2:0] unused_pad

);

parameter xuart_opt = 1'b1;

parameter can_opt = 1'b0;



parameter spi_opt = 1'b1;
parameter blockram_opt = 1'b1;
parameter sdcard_opt = 1'b1;

/****************************************************************************
 * TS-4700 boilerplate
 ****************************************************************************/

wire pll_100mhz, pll_locked, pll_75mhz, pll_300mhz;
reg [15:0] rst_counter = 16'd0;
reg rstn = 1'b0;
reg lock_q;
always @(posedge wb_clk) begin
  //lock_q <= pll_locked && pll2_locked && cpu_resetn_pad;
  lock_q <= pll_locked && cpu_resetn_pad;
  if (lock_q) rst_counter <= rst_counter + 1'b1;
  else begin
    rst_counter <= 16'd0;
    rstn <= 1'b0;
  end
  if (rst_counter[15]) rstn <= 1'b1;
end
wire wb_rst = !rstn;
pll1 ts4710pll (
  .CLK(fpga_25mhz_pad),
  .CLKOP(pll_300mhz),
  .CLKOK(pll_75mhz),
  .CLKOK2(pll_100mhz),
  .LOCK(pll_locked)
);

pll2 m0pll (
  .CLK(pll_300mhz),
  .CLKOP(pll_12mhz)
);

wire reboot;
assign en_usb_5v_pad = 1'b1;
assign reboot_pad = !reboot;  /* active low */

assign unused_pad = 3'h0;
reg [3:0] clk300mhz_cnt;
reg m0_clk;
assign m0_12mhz_clk_pad = pll_12mhz; // m0_clk;
always @(posedge pll_300mhz) begin
  clk300mhz_cnt <= clk300mhz_cnt + 1'b1;
  if (clk300mhz_cnt[3:0] == 12'd11 && m0_clk) begin
    m0_clk <= 1'b0;
    clk300mhz_cnt <= 4'd0;
  end
  if (clk300mhz_cnt[3:0] == 12'd12 && !m0_clk) begin
    m0_clk <= 1'b1;
    clk300mhz_cnt <= 4'd0;
  end
end

reg [31:0] usecs;
reg [6:0] count_1mhz;
always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    count_1mhz <= 7'd0;
    usecs <= 32'd0;
  end else begin
    count_1mhz <= count_1mhz + 1;
    if (count_1mhz == 7'd74) begin // 75MHz wb_clk
      count_1mhz <= 7'd0;
      usecs <= usecs + 1;
    end
  end
end

reg [2:0] count_14mhz; // actually 14.2857 MHz (100/7)
reg clk_14mhz;
always @(posedge pll_100mhz or posedge wb_rst) begin
  if (wb_rst) begin
    count_14mhz <= 3'd0;
    clk_14mhz <= 1'b0;
  end else begin
    count_14mhz <= count_14mhz + 1'b1;
    if (count_14mhz == 3'd6) count_14mhz <= 3'd0;
    if (count_14mhz == 3'd0) clk_14mhz <= 1'b0;
    if (count_14mhz == 3'd3) clk_14mhz <= 1'b1;
  end
end

assign db9_txd_pad = en_db9con ? debug_txd_pad : xuart_tx[4];
assign debug_rxd_pad = en_db9con ? (db9_rxd_pad & m0_txd_pad) : m0_txd_pad ;
assign m0_rxd_pad = debug_txd_pad;

/****************************************************************************
 * CPU bridge 
 ****************************************************************************/

wire smc_irq;
assign smc_irq_pad = smc_irq;

wire cpuwbm_cyc_o, cpuwbm_stb_o, cpuwbm_we_o;
wire [15:0] cpuwbm_dat_o;
wire [31:0] cpuwbm_adr, cpuwbm_adr_o, cpuwbm2_adr;
assign cpuwbm_adr_o = {15'd0, cpuwbm_adr[16:0]};
wire [14:0] ts4710_decodes = cpuwbm_adr[31:17];
wire [14:0] ts4710_decodes_100mhz = cpuwbm2_adr[31:17];
reg [15:0] cpuwbm_dat_i, cpuwbm2_dat_i;
wire [1:0] cpuwbm_sel_o;
reg cpuwbm_ack_i, cpuwbm2_ack_i;


wire [15:0] data_o;
wire data_oe;
assign smc_d_pad = data_oe ? data_o : 16'hzzzz;

wire smc_waitn;
assign smc_rdy_pad = smc_waitn;

/* Corrected notes on SMC_ADV# signal.  The built-in bootrom uses LLA#.  Once
 * we have control, we turn LLA# into a GPIO so that it is not used, and our
 * bus settings use #LUA.
 */

reg cpuwbm2_en;
wire cpuwbm2_cyc_o, cpuwbm2_stb_o, cpuwbm2_we_o;
wire [15:0] cpuwbm2_dat_o;
wire [31:0] cpuwbm2_adr_o, cpuwbs_adr_i;
assign cpuwbm2_adr_o = {15'd0, cpuwbm_adr[16:0]};
wire [1:0] cpuwbm2_sel_o;
wire [3:0] cpuwbm2_32_sel_o = 
  cpuwbm2_adr_o[1] ? {cpuwbm2_sel_o, 2'b00} : {2'b00, cpuwbm2_sel_o};
wire [15:0] cpuwbs_dat_i, cpuwbs_dat_o;
wire cpuwbs_cyc_i, cpuwbs_stb_i, cpuwbs_we_i, cpuwbs_ack_o;
wire [2:0] cpuwbs_cti_i;
wire [1:0] cpuwbs_bte_i;
mvsmcbridge #(
    
  .reset_smc_sync_en(1)


             
) cpu(
  .wb_rst_i(wb_rst),
  .wbm_clk_o(wb_clk),
  .wbm_cyc_o(cpuwbm_cyc_o),
  .wbm_stb_o(cpuwbm_stb_o),
  .wbm_dat_o(cpuwbm_dat_o),
  .wbm_adr_o(cpuwbm_adr),
  .wbm_we_o(cpuwbm_we_o),
  .wbm_dat_i(cpuwbm_dat_i),
  .wbm_ack_i(cpuwbm_ack_i),
  .wbm_sel_o(cpuwbm_sel_o),  

  /* WBM2 is at 100Mhz */
  .wbm2_en_i(cpuwbm2_en),
  .wbm2_clk_i(pll_100mhz),
  .wbm2_cyc_o(cpuwbm2_cyc_o),
  .wbm2_stb_o(cpuwbm2_stb_o),
  .wbm2_we_o(cpuwbm2_we_o),
  .wbm2_dat_o(cpuwbm2_dat_o),
  .wbm2_adr_o(cpuwbm2_adr),
  .wbm2_sel_o(cpuwbm2_sel_o),
  .wbm2_ack_i(cpuwbm2_ack_i),
  .wbm2_dat_i(cpuwbm2_dat_i),

  /* For satisfying read bursts */
  .wbs_clk_i(pll_100mhz),
  .wbs_cyc_i(cpuwbs_cyc_i),
  .wbs_stb_i(cpuwbs_stb_i),
  .wbs_we_i(cpuwbs_we_i),
  .wbs_adr_i(cpuwbs_adr_i),
  .wbs_dat_i(cpuwbs_dat_i),
  .wbs_dat_o(cpuwbs_dat_o),
  .wbs_cti_i(cpuwbs_cti_i),
  .wbs_bte_i(cpuwbs_bte_i),
  .wbs_ack_o(cpuwbs_ack_o),

  .smc_be1n_i(smc_be1n_pad),
  .smc_be2n_i(smc_be2n_pad),
  .smc_dat_i(smc_d_pad),
  .smc_dat_o(data_o),
  .smc_dat_oe_o(data_oe),
  .smc_rdn_i(smc_rdn_pad),
  .smc_wrn_i(smc_wen_pad),
  .smc_cs1n_i(smc_cs1n_pad),
  .smc_advn_i(smc_luan_pad & smc_llan_pad),
  .smc_waitn_o(smc_waitn),
  .smc_clk_i(smc_clk_pad),
  .clk_75mhz_i(pll_75mhz),

  .wr_fifo_empty_o(wr_fifo_empty)
);


/****************************************************************************
 * SD card + DMA core
 ****************************************************************************/
wire sd_power, sd_busy, sd_cmd, sd_cmd_oe, dmawbs_ack_o;
wire [7:0] sd_clk;
wire [31:0] sdwbs1_dat, dmawbs_dat;
wire [15:0] sdwbs1_dat_o = 
  cpuwbm2_adr_o[1] ? sdwbs1_dat[31:16] : sdwbs1_dat[15:0];
wire [31:0] sdwbs2_dat_o;
wire [15:0] dmawbs_dat_o =
  cpuwbm2_adr_o[1] ? dmawbs_dat[31:16] : dmawbs_dat[15:0];
wire sdwbs1_ack, sdwbs2_ack;
reg sdwbs1_en, sdwbs2_en, dmawbs_en;
wire [31:0] dmawbm2_adr_o, dmawbm2_dat_o;
wire dmawbm2_cyc_o, dmawbm2_stb_o, dmawbm2_we_o;
wire [3:0] dmawbm2_sel_o;
wire [2:0] dmawbm2_cti_o;
wire [1:0] dmawbm2_bte_o;
wire [3:0] sd_dat, sd_dat_oe;
wire [2:0] sd_lun;
wire sd_cmd_pad_in, sd_d0_pad_in, sd_d1_pad_in, sd_d2_pad_in, sd_d3_pad_in;
BB sdcmd(.I(sd_cmd), .T(!sd_cmd_oe), .B(sd_cmd_pad), .O(sd_cmd_pad_in));    
BB sdd0(.I(sd_dat[0]), .T(!sd_dat_oe[0]), .B(sd_d0_pad), .O(sd_d0_pad_in)); 
BB sdd1(.I(sd_dat[1]), .T(!sd_dat_oe[1]), .B(sd_d1_pad), .O(sd_d1_pad_in)); 
BB sdd2(.I(sd_dat[2]), .T(!sd_dat_oe[2]), .B(sd_d2_pad), .O(sd_d2_pad_in)); 
BB sdd3(.I(sd_dat[3]), .T(!sd_dat_oe[3]), .B(sd_d3_pad), .O(sd_d3_pad_in));
wire [59:0] dio_in;
assign sd_powern_pad = wb_rst ? 1'b0 : !sd_power;
assign sd_clk_fs_pad = sd_clk[2];
assign sd_clk_mic_pad = sd_clk[0];
assign emmc_resetn_pad = sd_power;
assign emmc_clk_pad = sd_clk[1];
wire emmc_cmd_pad_in, emmc_d0_pad_in, emmc_d1_pad_in, emmc_d2_pad_in, emmc_d3_pad_in;
BB emmccmd(.I(sd_cmd), .T(!sd_cmd_oe), .B(emmc_cmd_pad), .O(emmc_cmd_pad_in));    
BB emmcd0(.I(sd_dat[0]), .T(!sd_dat_oe[0]), .B(emmc_d_pad[0]), .O(emmc_d0_pad_in)); 
BB emmcd1(.I(sd_dat[1]), .T(!sd_dat_oe[1]), .B(emmc_d_pad[1]), .O(emmc_d1_pad_in)); 
BB emmcd2(.I(sd_dat[2]), .T(!sd_dat_oe[2]), .B(emmc_d_pad[2]), .O(emmc_d2_pad_in)); 
BB emmcd3(.I(sd_dat[3]), .T(!sd_dat_oe[3]), .B(emmc_d_pad[3]), .O(emmc_d3_pad_in));
assign emmc_d_pad[7:4] = sd_power ? 4'bzzzz : 4'b0000;
wire [3:0] sd_dat_in = sd_lun[1:0] == 2'd1 ? 
  {emmc_d3_pad_in, emmc_d2_pad_in, emmc_d1_pad_in, emmc_d0_pad_in} : 
  {sd_d3_pad_in, sd_d2_pad_in, sd_d1_pad_in, sd_d0_pad_in};
wire sd_cmd_in = sd_lun[1:0] == 2'd1 ? emmc_cmd_pad_in : sd_cmd_pad_in;
assign sd_act_ledn_pad = wr_fifo_empty && !sd_busy && 
  !(cpuwbm2_cyc_o && cpuwbm2_stb_o && sdwbs1_en) && !(dmawbm2_cyc_o && dmawbm2_stb_o);
ts_sdcore #(.multicard(1'b1)) sdcardcore(
  .wb_rst_i(wb_rst),
  .wb_clk_i(pll_100mhz & sdcard_opt),

  .wb1_adr_i(cpuwbm2_adr_o),
  .wb1_cyc_i(cpuwbm2_cyc_o && sdwbs1_en),
  .wb1_stb_i(cpuwbm2_stb_o && sdwbs1_en),
  .wb1_dat_o(sdwbs1_dat),
  .wb1_we_i(cpuwbm2_we_o),
  .wb1_dat_i({cpuwbm2_dat_o, cpuwbm2_dat_o}),
  .wb1_ack_o(sdwbs1_ack_o),
  .wb1_sel_i(cpuwbm2_32_sel_o),
  .wb1_bte_i(2'b00),
  .wb1_cti_i(3'b000),

  /* Connection to DMA core */
  .wb2_adr_i(32'h4 /* dmawbm2_adr_o */),
  .wb2_cyc_i(dmawbm2_cyc_o /* && sdwbs2_en */),
  .wb2_stb_i(dmawbm2_stb_o /* && sdwbs2_en */),
  .wb2_dat_o(sdwbs2_dat_o),
  .wb2_we_i(dmawbm2_we_o),
  .wb2_dat_i(dmawbm2_dat_o),
  .wb2_ack_o(sdwbs2_ack_o),
  .wb2_sel_i(4'b0011 /* dmawbm2_sel_o */),
  .wb2_bte_i(dmawbm2_bte_o),
  .wb2_cti_i(dmawbm2_cti_o),

  .sd_cmd_i(sd_cmd_in),
  .sd_cmd_o(sd_cmd),
  .sd_cmd_oe_o(sd_cmd_oe),
  .sd_dat_i(sd_dat_in),
  .sd_dat_o(sd_dat),
  .sd_dat_oe_o(sd_dat_oe),
  .sd_clk_o(sd_clk),

  .sd_wprot_i(8'b00000000),
  .sd_detect_i(8'b00000111),
  .sd_power_o(sd_power),
  .sd_lun_o(sd_lun),
  .sd_busy_o(sd_busy)
);

wire dmawbm1_dack, dmawbm2_dack;
wb_dma dmacore(
  .wb_clk_i(pll_100mhz),
  .wb_rst_i(wb_rst),

  /* wbs gets connected to 100MHz CPU wishbone */
  .wbs_cyc_i(cpuwbm2_cyc_o && dmawbs_en),
  .wbs_stb_i(cpuwbm2_stb_o && dmawbs_en),
  .wbs_we_i(cpuwbm2_we_o),
  .wbs_adr_i(cpuwbm2_adr_o),
  .wbs_dat_o(dmawbs_dat),
  .wbs_ack_o(dmawbs_ack_o),
  .wbs_dat_i({cpuwbm2_dat_o, cpuwbm2_dat_o}),
  .wbs_sel_i(cpuwbm2_32_sel_o),

  /* wbm1 gets connected to the SMC burst fifo wbs */
  .wbm1_cyc_o(cpuwbs_cyc_i),
  .wbm1_stb_o(cpuwbs_stb_i),
  .wbm1_adr_o(cpuwbs_adr_i),
  .wbm1_dat_o(cpuwbs_dat_i),
  .wbm1_dat_i(cpuwbs_dat_o),
  .wbm1_we_o(cpuwbs_we_i),
  .wbm1_ack_i(cpuwbs_ack_o),
  .wbm1_cti_o(cpuwbs_cti_i),
  .wbm1_bte_o(cpuwbs_bte_i),
  .wbm1_drq_i(1'b1), /* SMC fifo is always ready */
  .wbm1_dack_o(dmawbm1_dack),

  /* wbm2 gets connected to the SD card core */
  .wbm2_cyc_o(dmawbm2_cyc_o),
  .wbm2_stb_o(dmawbm2_stb_o),
  .wbm2_adr_o(dmawbm2_adr_o),
  .wbm2_dat_o(dmawbm2_dat_o),
  .wbm2_dat_i(sdwbs2_dat_o[15:0]),
  .wbm2_we_o(dmawbm2_we_o),
  .wbm2_ack_i(sdwbs2_ack_o),
  .wbm2_cti_o(dmawbm2_cti_o),
  .wbm2_bte_o(dmawbm2_bte_o),
  .wbm2_drq_i(1'b1),
  .wbm2_dack_o(dmawbm2_dack)
);


/****************************************************************************
 * SYSCON
 ****************************************************************************/

wire [59:0] dio;
wire [59:0] dio_oe;

wire [15:0] fpga_irqs;
assign fpga_irqs[0] = xuart_irq;
assign fpga_irqs[1] = !can_irqn;
assign fpga_irqs[3] = isa_irq5_pad;
assign fpga_irqs[4] = isa_irq6_pad;
assign fpga_irqs[5] = isa_irq7_pad;
assign fpga_irqs[6] = isa_irq9_pad;
assign fpga_irqs[7] = isa_irq3_pad;

wire green_led, red_led;
assign green_ledn_pad = green_led;
assign red_ledn_pad = red_led ? 1'b0 : 1'b1;
reg sd_busy_q, dmawbm1_dack_q, dmawbm2_dack_q;
always @(posedge wb_clk) begin
  sd_busy_q <= sd_busy;
  dmawbm1_dack_q <= dmawbm1_dack;
  dmawbm2_dack_q <= dmawbm2_dack;
end
reg scwbs_en;
wire scwbs_ack_o, adc_scl, adc_sda;
assign adc_clk_pad = adc_scl ? 1'bz : 1'b0;
assign adc_dat_pad = adc_sda ? 1'bz : 1'b0;
wire [15:0] scwbs_dat_o;
wire [11:0] lcd_oe, lcd;
wire [15:0] isa_dat, isa_dat_oe, isa_cn_d, isa_cn_d_oe;
wire [19:0] isa_add, isa_add_oe;
wire [24:0] isa_misc, isa_misc_oe, isa_misc_i;
wire isa_memory_en, sc_isa_memory_en, isa_dir, isa_data_en, isa_16_en;
wire isa_ts_special16_disable;
assign isa_add_pad[19] = xisa_add_oe[19] ? xisa_add[19] : 1'bz;
assign isa_add_pad[18] = xisa_add_oe[18] ? xisa_add[18] : 1'bz;
assign isa_add_pad[17] = xisa_add_oe[17] ? xisa_add[17] : 1'bz;
assign isa_add_pad[16] = xisa_add_oe[16] ? xisa_add[16] : 1'bz;
assign isa_add_pad[15] = xisa_add_oe[15] ? xisa_add[15] : 1'bz;
assign isa_add_pad[14] = xisa_add_oe[14] ? xisa_add[14] : 1'bz;
assign isa_add_pad[13] = xisa_add_oe[13] ? xisa_add[13] : 1'bz;
assign isa_add_pad[12] = xisa_add_oe[12] ? xisa_add[12] : 1'bz;
assign isa_add_pad[11] = xisa_add_oe[11] ? xisa_add[11] : 1'bz;
assign isa_add_pad[10] = xisa_add_oe[10] ? xisa_add[10] : 1'bz;
assign isa_add_pad[9] = xisa_add_oe[9] ? xisa_add[9] : 1'bz;
assign isa_add_pad[8] = xisa_add_oe[8] ? xisa_add[8] : 1'bz;
assign isa_add_pad[7] = xisa_add_oe[7] ? xisa_add[7] : 1'bz;
assign isa_add_pad[6] = xisa_add_oe[6] ? xisa_add[6] : 1'bz;
assign isa_add_pad[5] = xisa_add_oe[5] ? xisa_add[5] : 1'bz;
assign isa_add_pad[4] = xisa_add_oe[4] ? xisa_add[4] : 1'bz;
assign isa_add_pad[3] = xisa_add_oe[3] ? xisa_add[3] : 1'bz;
assign isa_add_pad[2] = xisa_add_oe[2] ? xisa_add[2] : 1'bz;
assign isa_add_pad[1] = xisa_add_oe[1] ? xisa_add[1] : 1'bz;
assign isa_add_pad[0] = xisa_add_oe[0] ? xisa_add[0] : 1'bz;

assign isa_dat_pad[15] = xisa_dat_oe[15] ? xisa_dat[15] : 1'bz;
assign isa_dat_pad[14] = xisa_dat_oe[14] ? xisa_dat[14] : 1'bz;
assign isa_dat_pad[13] = xisa_dat_oe[13] ? xisa_dat[13] : 1'bz;
assign isa_dat_pad[12] = xisa_dat_oe[12] ? xisa_dat[12] : 1'bz;
assign isa_dat_pad[11] = xisa_dat_oe[11] ? xisa_dat[11] : 1'bz;
assign isa_dat_pad[10] = xisa_dat_oe[10] ? xisa_dat[10] : 1'bz;
assign isa_dat_pad[9] = xisa_dat_oe[9] ? xisa_dat[9] : 1'bz;
assign isa_dat_pad[8] = xisa_dat_oe[8] ? xisa_dat[8] : 1'bz;
assign isa_dat_pad[7] = xisa_dat_oe[7] ? xisa_dat[7] : 1'bz;
assign isa_dat_pad[6] = xisa_dat_oe[6] ? xisa_dat[6] : 1'bz;
assign isa_dat_pad[5] = xisa_dat_oe[5] ? xisa_dat[5] : 1'bz;
assign isa_dat_pad[4] = xisa_dat_oe[4] ? xisa_dat[4] : 1'bz;
assign isa_dat_pad[3] = xisa_dat_oe[3] ? xisa_dat[3] : 1'bz;
assign isa_dat_pad[2] = xisa_dat_oe[2] ? xisa_dat[2] : 1'bz;
assign isa_dat_pad[1] = xisa_dat_oe[1] ? xisa_dat[1] : 1'bz;
assign isa_dat_pad[0] = xisa_dat_oe[0] ? xisa_dat[0] : 1'bz;

assign isa_cn_d_pad[15] = isa_cn_d_oe[15] ? isa_cn_d[15] : 1'bz;
assign isa_cn_d_pad[14] = isa_cn_d_oe[14] ? isa_cn_d[14] : 1'bz;
assign isa_cn_d_pad[13] = isa_cn_d_oe[13] ? isa_cn_d[13] : 1'bz;
assign isa_cn_d_pad[12] = isa_cn_d_oe[12] ? isa_cn_d[12] : 1'bz;
assign isa_cn_d_pad[11] = isa_cn_d_oe[11] ? isa_cn_d[11] : 1'bz;
assign isa_cn_d_pad[10] = isa_cn_d_oe[10] ? isa_cn_d[10] : 1'bz;
assign isa_cn_d_pad[9] = isa_cn_d_oe[9] ? isa_cn_d[9] : 1'bz;
assign isa_cn_d_pad[8] = isa_cn_d_oe[8] ? isa_cn_d[8] : 1'bz;
assign isa_cn_d_pad[7] = isa_cn_d_oe[7] ? isa_cn_d[7] : 1'bz;
assign isa_cn_d_pad[6] = isa_cn_d_oe[6] ? isa_cn_d[6] : 1'bz;
assign isa_cn_d_pad[5] = isa_cn_d_oe[5] ? isa_cn_d[5] : 1'bz;
assign isa_cn_d_pad[4] = isa_cn_d_oe[4] ? isa_cn_d[4] : 1'bz;
assign isa_cn_d_pad[3] = isa_cn_d_oe[3] ? isa_cn_d[3] : 1'bz;
assign isa_cn_d_pad[2] = isa_cn_d_oe[2] ? isa_cn_d[2] : 1'bz;
assign isa_cn_d_pad[1] = isa_cn_d_oe[1] ? isa_cn_d[1] : 1'bz;
assign isa_cn_d_pad[0] = 1'b0;

wire [7:0] ts_special16_dat = {
  isa_bale_pad, /* 24 */
  isa_tc_pad, /* 23 */
  isa_dack2_pad, /* 22 */
  isa_bclk_pad, /* 21 */
  isa_irq3_pad, /* 20 */
  isa_drq1_pad, /* 19 */
  isa_dack1_pad, /* 18 */
  isa_irq9_pad /* 17 */
};
assign isa_iorn_pad =	xisa_misc_oe[0] ? xisa_misc[0] : 1'bz;
assign isa_iown_pad =	xisa_misc_oe[1] ? xisa_misc[1] : 1'bz;
assign isa_memwn_pad =	xisa_misc_oe[2] ? xisa_misc[2] : 1'bz;
assign isa_memrn_pad =	xisa_misc_oe[3] ? xisa_misc[3] : 1'bz;
assign isa_reset_pad =	isa_misc_oe[4] ? isa_misc[4] : 1'bz;
assign isa_14mhz_pad =	isa_misc_oe[5] ? isa_misc[5] : clk_14mhz;
assign isa_endx_pad =	isa_misc_oe[6] ? isa_misc[6] : 1'bz;
assign isa_aen_pad =	isa_misc_oe[7] ? isa_misc[7] : 1'bz;
assign isa_refresh_pad= isa_misc_oe[8] ? isa_misc[8] : 1'bz;
assign isa_iochk_pad =	isa_misc_oe[9] ? isa_misc[9] : 1'bz;
assign isa_iordy_pad =	isa_misc_oe[10] ? isa_misc[10] : 1'bz;
assign isa_drq3_pad =	isa_misc_oe[11] ? isa_misc[11] : 1'bz;
assign isa_drq2_pad =	isa_misc_oe[12] ? isa_misc[12] : 1'bz;
assign isa_dack3_pad =	isa_misc_oe[13] ? isa_misc[13] : 1'bz;
assign isa_irq5_pad =	isa_misc_oe[14] ? isa_misc[14] : 1'bz;
assign isa_irq6_pad =	isa_misc_oe[15] ? isa_misc[15] : 1'bz;
assign isa_irq7_pad =	isa_misc_oe[16] ? isa_misc[16] : 1'bz;
assign isa_irq9_pad =	xisa_misc_oe[17] ? xisa_misc[17] : 1'bz;
assign isa_dack1_pad =	xisa_misc_oe[18] ? xisa_misc[18] : 1'bz;
assign isa_drq1_pad =	xisa_misc_oe[19] ? xisa_misc[19] : 1'bz;
assign isa_irq3_pad =	xisa_misc_oe[20] ? xisa_misc[20] : 1'bz;
assign isa_bclk_pad =	xisa_misc_oe[21] ? xisa_misc[21] : 1'bz;
assign isa_dack2_pad =	xisa_misc_oe[22] ? xisa_misc[22] : 1'bz;
assign isa_tc_pad =	xisa_misc_oe[23] ? xisa_misc[23] : 1'bz;
assign isa_bale_pad =	xisa_misc_oe[24] ? xisa_misc[24] : 1'bz;
assign isa_misc_i[0] = isa_iorn_pad;
assign isa_misc_i[1] = isa_iown_pad;
assign isa_misc_i[2] = isa_memwn_pad;
assign isa_misc_i[3] = isa_memrn_pad;
assign isa_misc_i[4] = isa_reset_pad;
assign isa_misc_i[5] = 1'b0;
assign isa_misc_i[6] = isa_endx_pad;
assign isa_misc_i[7] = isa_aen_pad;
assign isa_misc_i[8] = isa_refresh_pad;
assign isa_misc_i[9] = isa_iochk_pad;
assign isa_misc_i[10] = isa_iordy_pad;
assign isa_misc_i[11] = isa_drq3_pad;
assign isa_misc_i[12] = isa_drq2_pad;
assign isa_misc_i[13] = isa_dack3_pad;
assign isa_misc_i[14] = isa_irq5_pad;
assign isa_misc_i[15] = isa_irq6_pad;
assign isa_misc_i[16] = isa_irq7_pad;
assign isa_misc_i[17] = isa_irq9_pad;
assign isa_misc_i[18] = isa_dack1_pad;
assign isa_misc_i[19] = isa_drq1_pad;
assign isa_misc_i[20] = isa_irq3_pad;
assign isa_misc_i[21] = isa_bclk_pad;
assign isa_misc_i[22] = isa_dack2_pad;
assign isa_misc_i[23] = isa_tc_pad;
assign isa_misc_i[24] = isa_bale_pad;

assign isa_data_enn_pad= xisa_data_enn;
assign isa_dir_pad = xisa_dir;

assign dio_pad[0] = dio_oe[0] ? dio[0] : 1'bz;
assign dio_pad[1] = dio_oe[1] ? dio[1] : 1'bz;
assign dio_pad[2] = dio_oe[2] ? dio[2] : 1'bz;
assign dio_pad[3] = dio_oe[3] ? dio[3] : 1'bz;
assign dio_pad[4] = dio_oe[4] ? dio[4] : 1'bz;
assign dio_pad[5] = dio_oe[5] ? dio[5] : 1'bz;
assign dio_pad[6] = dio_oe[6] ? dio[6] : 1'bz;
assign dio_pad[7] = dio_oe[7] ? dio[7] : 1'bz;
assign dio_pad[8] = dio_oe[8] ? dio[8] : 1'bz;
assign en_dio_fet_pad = dio[9];
assign dio_in[8:0] = dio_pad[8:0];

assign lcd_en_pad = lcd_oe[0] ? lcd[0] : 1'bz;
assign lcd_wrn_pad = lcd_oe[1] ? lcd[1] : 1'bz;
assign lcd_rs_pad = lcd_oe[2] ? lcd[2] : 1'bz;
assign lcd_bias_pad = lcd_oe[3] ? lcd[3] : 1'bz;
assign lcd_pad[7] = lcd_oe[4] ? lcd[4] : 1'bz; 
assign lcd_pad[6] = lcd_oe[5] ? lcd[5] : 1'bz; 
assign lcd_pad[5] = lcd_oe[6] ? lcd[6] : 1'bz; 
assign lcd_pad[4] = lcd_oe[7] ? lcd[7] : 1'bz; 
assign lcd_pad[3] = lcd_oe[8] ? lcd[8] : 1'bz; 
assign lcd_pad[2] = lcd_oe[9] ? lcd[9] : 1'bz; 
assign lcd_pad[1] = lcd_oe[10] ? lcd[10] : 1'bz; 
assign lcd_pad[0] = lcd_oe[11] ? lcd[11] : 1'bz; 
wire [11:0] lcd_in;
assign lcd_in[11] = lcd_pad[0];
assign lcd_in[10] = lcd_pad[1];
assign lcd_in[9] = lcd_pad[2];
assign lcd_in[8] = lcd_pad[3];
assign lcd_in[7] = lcd_pad[4];
assign lcd_in[6] = lcd_pad[5];
assign lcd_in[5] = lcd_pad[6];
assign lcd_in[4] = lcd_pad[7];
assign lcd_in[3] = lcd_bias_pad;
assign lcd_in[2] = lcd_rs_pad;
assign lcd_in[1] = lcd_wrn_pad;
assign lcd_in[0] = lcd_en_pad;

wire m0_spare, m0_spare_oe, en_db9con;
assign m0_spare_pad = m0_spare_oe ? m0_spare : 1'bz;

wire scwbm_cyc, scwbm_stb, scwbm_we, scwbm_ack;
wire [15:0] scwbm_idat, scwbm_odat;
wire [31:0] scwbm_adr;
wire [1:0] sd_led, scwbm_sel;
assign fault_emmc_ledn_pad = !sd_led[0];
assign fault_sd_ledn_pad = !sd_led[1];
wb_buf scbuf(
  .wb_clk_i(wb_clk),
  .wb_rst_i(wb_rst),

  .wbs_cyc_i(cpuwbm_cyc_o && scwbs_en),
  .wbs_stb_i(cpuwbm_stb_o && scwbs_en),
  .wbs_we_i(cpuwbm_we_o),
  .wbs_adr_i(cpuwbm_adr_o),
  .wbs_dat_i(cpuwbm_dat_o),
  .wbs_sel_i(cpuwbm_sel_o),
  .wbs_dat_o(scwbs_dat_o),
  .wbs_ack_o(scwbs_ack_o),

  .wbm_cyc_o(scwbm_cyc),
  .wbm_stb_o(scwbm_stb),
  .wbm_adr_o(scwbm_adr),
  .wbm_dat_o(scwbm_odat),
  .wbm_dat_i(scwbm_idat),
  .wbm_ack_i(scwbm_ack),
  .wbm_sel_o(scwbm_sel),
  .wbm_we_o(scwbm_we)
);


syscon #(
    
  .wdog_default(3)


             
) sysconcore(
  .wb_clk_i(wb_clk),
  .wb_rst_i(wb_rst),

  .wb_cyc_i(scwbm_cyc),
  .wb_stb_i(scwbm_stb),
  .wb_we_i(scwbm_we),
  .wb_adr_i(scwbm_adr),
  .wb_dat_i(scwbm_odat),
  .wb_sel_i(scwbm_sel),
  .wb_dat_o(scwbm_idat),
  .wb_ack_o(scwbm_ack),

  .dio_o(dio),
  .dio_oe_o(dio_oe),
  .dio_i(dio_in),
  .usec_i(usecs),
  .mode1_i(mode1),
  .mode2_i(mode2),

  .red_led_o(red_led),
  .green_led_o(green_led),
  .reboot_o(reboot),
  .irqs_i(fpga_irqs),
  .irq_o(smc_irq),
  .sd_busy_i(sd_busy_q),
  .dma_idle_i(dmawbm1_dack_q && dmawbm2_dack_q),
  .adc_sda_i(adc_dat_pad),
  .adc_scl_i(adc_clk_pad),
  .adc_a_o(adc_a_pad),
  .adc_sda_o(adc_sda),
  .adc_scl_o(adc_scl),

  .lcd_o(lcd),
  .lcd_oe_o(lcd_oe),
  .lcd_i(lcd_in),

  .isa_add_o(isa_add),
  .isa_add_oe_o(isa_add_oe),
  .isa_add_i(isa_add_pad),
  .isa_dat_o(isa_dat),
  .isa_dat_oe_o(isa_dat_oe),
  .isa_dat_i(isa_dat_pad),
  .isa_misc_o(isa_misc),
  .isa_misc_oe_o(isa_misc_oe),
  .isa_misc_i(isa_misc_i),
  .isa_cn_d_o(isa_cn_d),
  .isa_cn_d_oe_o(isa_cn_d_oe),
  .isa_cn_d_i(isa_cn_d_pad),
  .isa_data_en_o(isa_data_en),
  .isa_dir_o(isa_dir),
  .isa_memory_en_o(sc_isa_memory_en),
  .isa_16_en_o(isa_16_en),
  .isa_ts_special16_disable_o(isa_ts_special16_disable),
  
  .db9_cts_i(db9_cts_pad),
  .db9_dsr_i(db9_dsr_pad),
  .db9_dcd_i(db9_dcd_pad),
  .com2_cts_i(com2_cts_pad),
  .com3_cts_i(com3_cts_pad),
  .db9_rts_o(db9_rts_pad),
  .db9_dtr_o(db9_dtr_pad),
  .com2_rts_o(com2_rts_pad),
  .com3_rts_o(com3_rts_pad),

  .m0_spare_o(m0_spare),
  .m0_spare_i(m0_spare_pad),
  .m0_spare_oe_o(m0_spare_oe),
  .sd_led_o(sd_led),
  .en_db9con_o(en_db9con),

  .jp1_i(jp1_pad),
  .jp2n_i(jp2n_pad)
);


/****************************************************************************
 * 16 KByte WISHBONE blockram 
 ****************************************************************************/

wire [31:0] sramwbs1_dat;
wire [15:0] sramwbs1_dat_o = 
  cpuwbm_adr_o[1] ? sramwbs1_dat[31:16] : sramwbs1_dat[15:0];
wire [3:0] cpuwbm32_sel_o =
  cpuwbm_adr_o[1] ? {cpuwbm_sel_o, 2'b00} : {2'b00, cpuwbm_sel_o};
wire sramwbs1_ack_o;
reg sramwbs1_en;
wire xuartwbm_cyc_o, xuartwbm_stb_o, xuartwbm_we_o, xuartwbm_ack_i;
wire [15:0] xuartwbm_adr_o, xuartwbm_dat_o, xuartwbm_dat_i;
wire [31:0] xuartwbm_dat32_i;
wire [3:0] xuartwbm_sel_o;
assign xuartwbm_dat_i = xuartwbm_adr_o[1] ?
  xuartwbm_dat32_i[31:16] : xuartwbm_dat32_i[15:0];
wire sramwbm1_cyc, sramwbm1_stb, sramwbm1_ack;
wire [31:0] sramwbm1_dat;
wb_readbuf srambuf(
  .wb_clk_i(wb_clk & blockram_opt),
  .wb_rst_i(wb_rst | ~blockram_opt),

  .wbs_cyc_i(cpuwbm_cyc_o && sramwbs1_en),
  .wbs_stb_i(cpuwbm_stb_o && sramwbs1_en),
  .wbs_we_i(cpuwbm_we_o),
  .wbs_dat_o(sramwbs1_dat),
  .wbs_ack_o(sramwbs1_ack_o),

  .wbm_cyc_o(sramwbm1_cyc),
  .wbm_stb_o(sramwbm1_stb),
  .wbm_ack_i(sramwbm1_ack),
  .wbm_dat_i(sramwbm1_dat)
);
blockram sramcore(
  .wb_clk_i(wb_clk & blockram_opt),
  .wb_rst_i(wb_rst | ~blockram_opt),

  .wb2_adr_i(cpuwbm_adr_o),
  .wb2_cyc_i(sramwbm1_cyc),
  .wb2_stb_i(sramwbm1_stb),
  .wb2_we_i(cpuwbm_we_o), 
  .wb2_sel_i(cpuwbm32_sel_o),
  .wb2_dat_i({cpuwbm_dat_o, cpuwbm_dat_o}),
  .wb2_dat_o(sramwbm1_dat),
  .wb2_ack_o(sramwbm1_ack),

  .wb3_cyc_i(xuartwbm_cyc_o),
  .wb3_stb_i(xuartwbm_stb_o),
  .wb3_adr_i(xuartwbm_adr_o),
  .wb3_dat_i({xuartwbm_dat_o, xuartwbm_dat_o}),
  .wb3_sel_i(xuartwbm_sel_o),
  .wb3_we_i(xuartwbm_we_o),
  .wb3_ack_o(xuartwbm_ack_i),
  .wb3_dat_o(xuartwbm_dat32_i)

);

/****************************************************************************
 * XUART core
 ****************************************************************************/

wire xuartwbs_ack;
wire xuartwbs_ack_o = xuartwbs_ack | !xuart_opt;
reg xuwbs_en;
wire [15:0] xuartwbs_dat_o;
wire [7:0] xuart_rx;
wire [7:0] xuart_active;
wire [7:0] xuart_cts = 8'hff;
wire [7:0] xuart_txen, xuart_tx;
reg [7:0] xuart_txen_q, xuart_tx_q;
reg [7:0] xuart_rx_q, xuart_cts_q;
wire xuart_irq;
wire xwbm_cyc_o, xwbm_stb_o, xwbm_we_o, xwbm_ack_i;
wire [15:0] xwbm_adr_o, xwbm_dat_o, xwbm_dat_i;
wire [3:0] xwbm_sel32_o = xwbm_adr_o[1] ? 4'b1100 : 4'b0011;
wire [31:0] xwbm_dat32_i;
wire [3:0] xwbm_sel_o;
assign xuart_rx[0] = com2_rxd_pad;
assign xuart_rx[1] = com3_rxd_pad;
assign xuart_rx[2] = rxd_485_pad | xuart_tx[2];
assign xuart_rx[3] = rxd_422_pad;
assign xuart_rx[4] = db9_rxd_pad;
assign com2_txd_pad = xuart_tx[0];
assign com3_txd_pad = xuart_tx[1];
assign txd_485_pad = xuart_tx[2] & xuart_tx[3];
assign tx_en_485_pad = xuart_txen[2] | xuart_txen[3];

always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    xuart_rx_q <= 8'hff;
    xuart_cts_q <= 8'hff;
    xuart_tx_q <= 8'hff;
    xuart_txen_q <= 8'hff;
  end else begin
    xuart_rx_q <= xuart_rx;
    xuart_cts_q <= xuart_cts;
    xuart_tx_q <= xuart_tx;
    xuart_txen_q <= xuart_txen;
  end
end
wb_buf xubuf(
  .wb_clk_i(wb_clk),
  .wb_rst_i(wb_rst),

  .wbs_cyc_i(xwbm_cyc_o),
  .wbs_stb_i(xwbm_stb_o),
  .wbs_we_i(xwbm_we_o),
  .wbs_dat_o(xwbm_dat_i),
  .wbs_dat_i(xwbm_dat_o),
  .wbs_ack_o(xwbm_ack_i),
  .wbs_adr_i(xwbm_adr_o),
  .wbs_sel_i(xwbm_sel32_o),

  .wbm_adr_o(xuartwbm_adr_o),
  .wbm_cyc_o(xuartwbm_cyc_o),
  .wbm_stb_o(xuartwbm_stb_o),
  .wbm_we_o(xuartwbm_we_o),
  .wbm_dat_o(xuartwbm_dat_o),
  .wbm_ack_i(xuartwbm_ack_i),
  .wbm_sel_o(xuartwbm_sel_o),
  .wbm_dat_i(xuartwbm_dat_i)
);
ts_xuart tsxuartcore(
  .wb_clk_i(wb_clk & xuart_opt),
  .wb_rst_i(wb_rst | !xuart_opt),

  .wbm_cyc_o(xwbm_cyc_o),
  .wbm_stb_o(xwbm_stb_o),
  .wbm_we_o(xwbm_we_o),
  .wbm_ack_i(xwbm_ack_i),
  .wbm_adr_o(xwbm_adr_o),
  .wbm_dat_i(xwbm_dat_i),
  .wbm_dat_o(xwbm_dat_o),

  .wbs_cyc_i(cpuwbm_cyc_o && xuwbs_en),
  .wbs_stb_i(cpuwbm_stb_o && xuwbs_en),
  .wbs_we_i(cpuwbm_we_o),
  .wbs_adr_i(cpuwbm_adr_o),
  .wbs_dat_i(cpuwbm_dat_o),
  .wbs_dat_o(xuartwbs_dat_o),
  .wbs_ack_o(xuartwbs_ack),

  .baudclk_i(pll_100mhz),
  .cts_i(xuart_cts_q),
  .rx_i(xuart_rx_q),
  .tx_o(xuart_tx),
  .txen_o(xuart_txen),
  .active_o(xuart_active),
  .irq_o(xuart_irq),
  .uart_disable_i(8'he0) 
);


/****************************************************************************
 * CAN core
 ****************************************************************************/

reg [4:0] seq_24mhz;
reg clk_24mhz;
always @(posedge pll_100mhz or posedge wb_rst) begin
  if (wb_rst) begin
    clk_24mhz <= 1'b1;
    seq_24mhz <= 5'd0;
  end else begin
    seq_24mhz <= seq_24mhz + 1'b1;
    case (seq_24mhz)
    1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21: clk_24mhz <= !clk_24mhz;
    24: begin
      clk_24mhz <= !clk_24mhz;
      seq_24mhz <= 5'd0;
    end
    endcase
  end
end

reg canwbs_en;
wire canwbs_ack;
wire canwbs_ack_o = canwbs_ack | !can_opt;
wire can_tx, can_irqn;
wire [7:0] canwbs_dat_i;
assign canwbs_dat_i = cpuwbm_sel_o[0] ? cpuwbm_dat_o[7:0] : cpuwbm_dat_o[15:8];
wire [7:0] canwbs_dat8_o;
wire [15:0] canwbs_dat_o = {canwbs_dat8_o, canwbs_dat8_o};
wire [7:0] canwbs_adr_i;
assign canwbs_adr_i[7:1] = cpuwbm_adr_o[7:1];
assign canwbs_adr_i[0] = !cpuwbm_sel_o[0];

can_top cancore(
  .wb_clk_i(wb_clk && can_opt),
  .wb_rst_i(wb_rst | !can_opt),
  .wb_cyc_i(cpuwbm_cyc_o && canwbs_en),
  .wb_stb_i(cpuwbm_stb_o && canwbs_en),
  .wb_we_i(cpuwbm_we_o),
  .wb_dat_i(canwbs_dat_i),
  .wb_dat_o(canwbs_dat8_o),
  .wb_adr_i(canwbs_adr_i),
  .wb_ack_o(canwbs_ack),

  .clk_i(clk_24mhz && can_opt),
  .rx_i(rxd_can_pad),
  .tx_o(txd_can_pad),
  .irq_on(can_irqn)
  //.bus_off_on()
);


/****************************************************************************
 * SPI core
 ****************************************************************************/

reg spiwbs_en;
wire [15:0] spiwbs_dat_o;
wire spiwbs_ack;
wire spiwbs_ack_o = spiwbs_ack | !spi_opt;
wire scken;
wire [3:0] spi_csn;
reg [3:0] spi_csn_q;
assign spi_csn_pad = spi_csn[0];
always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) spi_csn_q <= 4'hf;
  else spi_csn_q <= spi_csn;
end
assign spi_miso_pad = 1'bz;
wire [3:0] spi_si = {4{spi_miso_pad}};
wire spi_sck, spi_clk;
assign spi_clk_pad = spi_sck;
wire spi_so;
wb_spi spicore(
  .wb_rst_i(wb_rst | !spi_opt),
  .wb_clk_i(wb_clk & spi_opt),
  .wb_cyc_i(cpuwbm_cyc_o && spiwbs_en),
  .wb_stb_i(cpuwbm_stb_o && spiwbs_en),
  .wb_dat_o(spiwbs_dat_o),
  .wb_dat_i(cpuwbm_dat_o),
  .wb_adr_i(cpuwbm_adr_o),
  .wb_we_i(cpuwbm_we_o),
  .wb_ack_o(spiwbs_ack),
  .wb_sel_i(cpuwbm_sel_o),

  .scken_o(scken),
  .so_o(spi_mosi_pad),
  .si_i(spi_si),
  .csn_o(spi_csn),
  .sck_o(spi_sck)
);

reg buswbs_en;
reg [15:0] isadat, isaadd, buswbs_dat;
reg iow, ior, buswbs_ack, isaoe;
reg [2:0] bus_state /* synthesis syn_encoding="original" */;
reg [4:0] bus_cntdown;
reg [24:0] xisa_misc, xisa_misc_oe;
reg xisa_dir, xisa_data_enn;
reg [15:0] xisa_dat_oe, xisa_dat;
reg [19:0] xisa_add, xisa_add_oe;
assign isa_memory_en = isaadd[15] && sc_isa_memory_en;
reg isa_special16_override;
always @(*) begin
  isa_special16_override = 1'b0;
  xisa_misc[24:0] = isa_misc[24:0];
  xisa_misc_oe[24:0] = isa_misc_oe[24:0];
  xisa_dir = isa_dir;
  xisa_data_enn = isa_data_en;
  xisa_add_oe[19:0] = isa_add_oe[19:0];
  xisa_add[19:0] = isa_add[19:0];
  xisa_dat_oe[15:0] = isa_dat_oe[15:0];
  xisa_dat[15:0] = isa_dat[15:0];
  if (bus_state != 3'd0) begin
    xisa_misc_oe[3:0] = 4'b1111;
    xisa_misc[3:0] = 4'b1111;
    if (isa_16_en) begin
      xisa_add[14:0] = isaadd[14:0];
      xisa_add_oe[14:0] = 15'h7fff;
      xisa_dat_oe[15:0] = 16'hffff;
      isa_special16_override = !isa_ts_special16_disable;;
    end else begin
      xisa_add[13:0] = isaadd[14:1];
      xisa_add_oe[13:0] = 14'h3fff;
      xisa_dat_oe[7:0] = 8'hff;
    end
    if (!isa_memory_en) begin
      xisa_add[19:12] = 8'd0;
      xisa_add_oe[19:12] = 8'b11111111;
    end
  end
  if (!isa_memory_en && ior) begin /* IO read */
    xisa_misc[0] = 1'b0;
    xisa_misc_oe[0] = 1'b1;
    xisa_dir = 1'b0; /* B to A */
    xisa_data_enn = 1'b0;
    isa_special16_override = isa_16_en && !isa_ts_special16_disable;
    if (isa_16_en) xisa_dat_oe[15:0] = 16'd0;
    else xisa_dat_oe[7:0] = 8'd0;
  end
  if (isa_memory_en && ior) begin /* mem read */
    xisa_misc[3] = 1'b0;
    xisa_misc_oe[3] = 1'b1;
    xisa_dir = 1'b0; /* B to A */
    xisa_data_enn = 1'b0;
    xisa_dat_oe[15:0] = 16'd0;
    isa_special16_override = isa_16_en && !isa_ts_special16_disable;
    if (isa_16_en) xisa_dat_oe[15:0] = 16'd0;
    else xisa_dat_oe[7:0] = 8'd0;
  end
  if (!isa_memory_en && iow) begin /* IO write */
    xisa_misc[1] = 1'b0;
    xisa_misc_oe[1] = 1'b1;
    xisa_dir = 1'b1; /* A to B */
    xisa_data_enn = 1'b0;
  end
  if (isa_memory_en && iow) begin /* mem write */
    xisa_misc[2] = 1'b0;
    xisa_misc_oe[2] = 1'b1;
    xisa_dir = 1'b1; /* A to B */
    xisa_data_enn = 1'b0;
  end
  if (isaoe) begin
    if (isa_16_en) begin
      isa_special16_override = !isa_ts_special16_disable;
      xisa_dat_oe[15:0] = 16'hffff;
      xisa_dat[15:0] = isadat[15:0];
    end else begin
      xisa_dat_oe[7:0] = 8'hff;
      xisa_dat[7:0] = isadat[7:0];
    end
  end

  if (isa_special16_override) begin
    xisa_misc_oe[24:17] = xisa_dat_oe[15:8];
    xisa_misc[24:17] = xisa_dat[15:8];
  end
end

always @(posedge wb_clk or posedge wb_rst) begin
  if (wb_rst) begin
    iow <= 1'b0;
    ior <= 1'b0;
    isaadd <= 16'd0;
    isadat <= 16'd0;
    isaoe <= 1'b0;
    buswbs_ack <= 1'b0;
    buswbs_dat <= 16'd0;
    bus_state <= 3'd0;
    bus_cntdown <= 5'b11111;
  end else begin
    buswbs_ack <= 1'b0;
    case (bus_state) 
    3'd0: if (buswbs_en && cpuwbm_cyc_o && cpuwbm_stb_o) begin
      bus_state <= 3'd1;
      isaadd[15:0] <= cpuwbm_adr_o[15:0];
      isadat[15:8] <= cpuwbm_dat_o[15:8];
      isadat[7] <= cpuwbm_dat_o[7];
      isadat[6] <= cpuwbm_dat_o[6];
      isadat[5] <= cpuwbm_dat_o[4];
      isadat[4] <= cpuwbm_dat_o[3];
      isadat[3] <= cpuwbm_dat_o[2];
      isadat[2] <= cpuwbm_dat_o[5];
      isadat[1] <= cpuwbm_dat_o[1];
      isadat[0] <= cpuwbm_dat_o[0];
      if (cpuwbm_sel_o == 2'b10) begin
        isaadd[0] <= 1'b1;
        isadat[7] <= cpuwbm_dat_o[15];
        isadat[6] <= cpuwbm_dat_o[14];
        isadat[5] <= cpuwbm_dat_o[12];
        isadat[4] <= cpuwbm_dat_o[11];
        isadat[3] <= cpuwbm_dat_o[10];
        isadat[2] <= cpuwbm_dat_o[13];
        isadat[1] <= cpuwbm_dat_o[9];
        isadat[0] <= cpuwbm_dat_o[8];
      end
      if (cpuwbm_we_o) isaoe <= 1'b1;
      bus_state <= 3'd1;
      bus_cntdown <= 5'd24;
    end
    3'd1: begin
      if (isaoe) iow <= 1'b1; else ior <= 1'b1;
      if (bus_cntdown == 5'd0) bus_state <= 3'd3;
      bus_cntdown <= bus_cntdown - 1'b1;
    end
    3'd3: begin
      iow <= 1'b0;
      ior <= 1'b0;
      buswbs_dat[15:8] <= isa_ts_special16_disable ? isa_dat_pad[15:8] : ts_special16_dat[7:0];
      buswbs_dat[7] <= isa_dat_pad[7];
      buswbs_dat[6] <= isa_dat_pad[6];
      buswbs_dat[5] <= isa_dat_pad[2];
      buswbs_dat[4] <= isa_dat_pad[5];
      buswbs_dat[3] <= isa_dat_pad[4];
      buswbs_dat[2] <= isa_dat_pad[3];
      buswbs_dat[1] <= isa_dat_pad[1];
      buswbs_dat[0] <= isa_dat_pad[0];
      if (isaadd[0]) begin
        buswbs_dat[15] <= isa_dat_pad[7];
        buswbs_dat[14] <= isa_dat_pad[6];
        buswbs_dat[13] <= isa_dat_pad[2];
        buswbs_dat[12] <= isa_dat_pad[5];
        buswbs_dat[11] <= isa_dat_pad[4];
        buswbs_dat[10] <= isa_dat_pad[3];
        buswbs_dat[9] <= isa_dat_pad[1];
        buswbs_dat[8] <= isa_dat_pad[0];
      end
      buswbs_ack <= 1'b1;
      bus_state <= 3'd2;
      bus_cntdown <= 5'd3;
    end
    3'd2: begin
      bus_cntdown <= bus_cntdown - 1'b1;
      if (bus_cntdown == 5'd0) begin
        bus_state <= 3'd0;
        isaoe <= 1'b0;
      end
    end
    endcase
  end
end



/****************************************************************************
 * Master address decode
 ****************************************************************************/

/*

64 KB FPGA address space:
start   stop    usage
-----   ----    -----
0x7000  0x7FFF  PC104 io bus access
0x8000  0xFFFF  PC104 mem bus access
0x5400          XUART IO registers
0x5000          touchscreen core
0x4D00  0x4DFF  2nd CAN core
0x4C00  0x4CFF  CAN core
0x4800          SPI core
0x4400          ADC core
0x4000          syscon registers
0x0000  0x3FFF  16KB blockram access

*/
always @(*) begin
  
  sramwbs1_en = ts4710_decodes[0];
  buswbs_en = ts4710_decodes[1];
  xuwbs_en = ts4710_decodes[2];
  //tswbs_en = ts4710_decodes[3];
  canwbs_en = ts4710_decodes[4];
  spiwbs_en = ts4710_decodes[5];
  scwbs_en = ts4710_decodes[7];
  sdwbs1_en = ts4710_decodes_100mhz[8];
  dmawbs_en = ts4710_decodes_100mhz[9];
  //sdwbs2_en = 1'b0;

  cpuwbm_dat_i = 16'hxxxx;
  cpuwbm_ack_i = cpuwbm_cyc_o && cpuwbm_stb_o;

  cpuwbm2_dat_i = 16'hxxxx;
  cpuwbm2_ack_i = cpuwbm2_cyc_o && cpuwbm2_stb_o;
  cpuwbm2_en = ts4710_decodes[8] | ts4710_decodes[9];

  if (ts4710_decodes[0]) begin
    cpuwbm_dat_i = sramwbs1_dat_o;
    cpuwbm_ack_i = sramwbs1_ack_o;
  end else if (ts4710_decodes[1]) begin
    cpuwbm_dat_i = buswbs_dat;
    cpuwbm_ack_i = buswbs_ack;
  end else if (ts4710_decodes[2]) begin
    cpuwbm_dat_i = xuartwbs_dat_o;
    cpuwbm_ack_i = xuartwbs_ack_o;
  //end else if (ts4710_decodes[3]) begin
    //cpuwbm_dat_i = tswbs_dat_o;
    //cpuwbm_ack_i = tswbs_ack_o;
  end else if (ts4710_decodes[4]) begin
    cpuwbm_dat_i = canwbs_dat_o;
    cpuwbm_ack_i = canwbs_ack_o;
  end else if (ts4710_decodes[5]) begin
    cpuwbm_dat_i = spiwbs_dat_o;
    cpuwbm_ack_i = spiwbs_ack_o;
  //end else if (ts4710_decodes[6]) begin
    //cpuwbm_dat_i = adcwbs_dat_o;
    //cpuwbm_ack_i = adcwbs_ack_o;
  end else if (ts4710_decodes[7]) begin
    cpuwbm_dat_i = scwbs_dat_o;
    cpuwbm_ack_i = scwbs_ack_o;
  end 

  if (ts4710_decodes_100mhz[8]) begin
    cpuwbm2_dat_i = sdwbs1_dat_o;
    cpuwbm2_ack_i = sdwbs1_ack_o;
  end else if (ts4710_decodes_100mhz[9]) begin
    cpuwbm2_dat_i = dmawbs_dat_o;
    cpuwbm2_ack_i = dmawbs_ack_o;
  end
end


endmodule
